
#-----redlWorld-----

setClass("redlWorld", representation(ref="externalptr",
	stateEnv="environment"))
setGeneric("getStatus", function(x)standardGeneric("getStatus"))
setMethod("getStatus", "redlWorld", function(x) get("status", x@stateEnv))
setGeneric("setStatus", function(x,y)standardGeneric("setStatus"))
setMethod("setStatus", c("redlWorld", "ANY"), function(x,y) 
  assign("status", y, x@stateEnv))
setGeneric("ref", function(x)standardGeneric("ref"))
setMethod("ref", "redlWorld", function(x) x@ref)

setMethod("show", "redlWorld", function(object) {
 cat("redlWorld object with status", getStatus(object),
	"\n")})

#-----redlURI-----

setClass("redlURI", representation(ref="externalptr",
	URIstring="character", world="redlWorld"))
setGeneric("URIstring", function(x)standardGeneric("URIstring"))
setMethod("URIstring", "redlURI", function(x) x@URIstring)
setMethod("show", "redlURI", function(object) {
 cat("redlURI object with URIstring:\n", URIstring(object), "\n") })
setMethod("ref", "redlURI", function(x) x@ref)

#-----redlModel-----

setClass("redlModel", representation(ref="externalptr",
	storagetype="character", stateEnv="environment",
        world="redlWorld"))
setMethod("ref", "redlModel", function(x) x@ref)
setMethod("show", "redlModel", function(object) {
 if (getStatus(object) != "open") 
	{
	cat("Non-open (freed) redlModel instance.  Nothing to show.\n")
	}
 else {
	cat("redlModel object", ", status=",
	  getStatus(object),".\n", sep="") 
	}
 })
setAs("redlModel", "data.frame", function(from,to) redlMod2df(from) )
setMethod("getStatus", "redlModel", function(x) get("status", x@stateEnv))
setMethod("setStatus", c("redlModel", "ANY"), function(x,y) 
  assign("status", y, x@stateEnv))

#-- redlNode

setClass("redlNode", representation(ref="externalptr",
        world="redlWorld"))
setMethod("ref", "redlNode", function(x) x@ref)
setGeneric("world", function(x) standardGeneric("world"))
setMethod("world", "redlNode", function(x) function(x) x@world)
setMethod("show", "redlNode", function(object) {
	cat("redl Node object in world\n")
	world(object)
 })

#-----operations

checkRedlWorld <- function(x) {
	if (getStatus(x) != "open") stop("status of RDF world not 'open'")
	invisible(TRUE)
}

openRedlWorld <- function(.force=FALSE) {
        if (!.force && exists("..GredlWorld")) {
                warning(
                   paste("global redland RDF world found, so no action taken.\n",
                   "Set .force to attempt to create a new RDF world.\n",sep=""))
                return(invisible(NULL))
        }
        ptr <- .Call("get_open_redl_world", PACKAGE="Rredland" )
	openv = new.env()
	assign("status", "open", openv)
	new("redlWorld", ref=ptr, stateEnv=openv)
}

setGeneric("freeRedl", function(x) standardGeneric("freeRedl"))
setMethod("freeRedl", "redlWorld", function(x) {
	checkRedlWorld(x)
        .Call("free_redl_world", ref(x), PACKAGE="Rredland" )
	setStatus(x, "free")
	invisible(NULL)
	})

setMethod("freeRedl", "redlModel", function(x) {
	checkRedlWorld(x)
        .Call("free_redl_model", ref(x), PACKAGE="Rredland" )
	setStatus(x, "free")
	invisible(NULL)
	})

setGeneric("makeRedlURI", function(uri,w) standardGeneric("makeRedlURI"))
setMethod("makeRedlURI", c("character", "redlWorld"), function(uri,w) {
 if (length(uri) > 1) stop("uri must have length 1")
 checkRedlWorld(w)
 rf <- .Call("get_redl_world_uri", ref(w), uri, PACKAGE="Rredland" )
 new("redlURI", ref=rf, world=w, URIstring=uri)
})
setMethod("makeRedlURI", c("character", "missing"), function(uri,w) {
 checkRedlWorld(..GredlWorld)
 makeRedlURI(uri, ..GredlWorld)
})

# stipulation for reading: we will create new storage as
# requested by the user, who can specify a redland hash for
# fine level of control
#
# if a non-null stoHash is supplied, it overrides the storageType

readRDF <- function(uri, storageType=c("internal", "bdb")[1], storageName="test",
	world=..GredlWorld, stoHash=NULL) {
	checkRedlWorld(world)
	if (is(uri, "character")) uri = makeRedlURI(uri, world)
	uriref = ref(uri)
	if (!(storageType %in% c("internal", "bdb"))) stop(
		"storageType must be either internal or bdb")
	stocode <- c("memory", "hashes")[which(storageType==c("internal", "bdb"))]
	if (is.null(stoHash) && storageType == "internal") stoHash=""
	else if (is.null(stoHash) && storageType == "bdb") 
		stoHash="new='yes',hash-type='bdb',dir='.'"
	storef = .Call("call_new_storage", ref(world), stocode, 
		storageName, stoHash, PACKAGE="Rredland" )
	modref = .Call("get_redl_model", ref(world), storef, PACKAGE="Rredland")
	parsref = .Call("get_redl_default_parser", ref(world),
		PACKAGE="Rredland")
	.Call("parse_into_redl_model", parsref, uriref, modref, PACKAGE="Rredland")
	tmp <- new("redlModel", ref=modref, storagetype=storageType, world=world)
	setStatus(tmp,"open")
	tmp
}

restoreBDB <- function(storageName, world=..GredlWorld, 
		stoHash="hash-type='bdb',dir='.'") {
	checkRedlWorld(world)
	storef = .Call("call_new_storage", ref(world), "hashes",
		storageName, stoHash, PACKAGE="Rredland")
	modref = .Call("get_redl_model", ref(world), storef, PACKAGE="Rredland")
	tmp <- new("redlModel", ref=modref, storagetype="bdb", world=world)
	setStatus(tmp, "open")
	tmp
}

getModelAsStream <- function(modRef) {
        if (is.null(modRef)) stop("null modRef")
        .Call("get_redl_model_as_stream", modRef, PACKAGE="Rredland")
}

popStatementFromStream <- function(streamRef, nullOnNull=TRUE) {
        if (is.null(streamRef))
                {
                if(!nullOnNull) stop("null streamRef but nullOnNull eq FALSE")
                else NULL
                }
        .Call("pop_statement_from_stream", streamRef, PACKAGE="Rredland")
}

getRedlModelSize <- function(modRef) {
        if (is.null(modRef)) stop("null modRef")
        att <- .Call("get_redl_model_size", modRef, PACKAGE="Rredland" )
        if (att < 0) {
		cat("Getting size by traversing stream.\n")
                str <- getModelAsStream(modRef)
                i <- 0
                while ( !is.null(tmp <- popStatementFromStream(str) ) ) {
                        i <- i+1
                        }
                return(i)
                }
        return(att)
}

setGeneric("computeSize", function(x) standardGeneric("computeSize"))
setMethod("computeSize", "redlModel", function(x) {
	getRedlModelSize(ref(x))
	})
#setGeneric("sizeEnv", function(x)standardGeneric("sizeEnv"))
#setMethod("sizeEnv", "redlModel", function(x) x@sizeEnv)
setGeneric("size", function(x) standardGeneric("size"))
setMethod("size", "redlModel", function(x) {
#	if (length(ls(sizeEnv(x))) == 0) 
#		{
#		tmp <- computeSize(x)
#		cat("Caching size in model object.\n")
#		assign("size", tmp, env=sizeEnv(x))
#		return(tmp)
#		}
#	else get("size", env=sizeEnv(x))
	computeSize(x)
	})


cleanXSDT <- function(x) gsub("\\^\\^.*>","", x)

redlMod2df <- function(mod) {
	if (getStatus(mod) != "open") stop("need to use an open redlModel instance")
	cleanRedlBrack <- function(x) gsub("\\[", "", gsub("\\]", "", gsub("\\}", "", gsub("\\{", "", x))))


        if (is(mod,"redlModel")) modref <- ref(mod)
        if (is.null(modref)) stop("null model reference pointer...")
        sz <- size(mod)
        str <- getModelAsStream(modref)
        out <- matrix("NA", nr=sz, nc=3)
        colnames(out) <- c("subject", "predicate", "object")
        i <- 0
        while ( !is.null(tmp <- popStatementFromStream(str) ) ) {
                i <- i+1
                tmp <- strsplit( cleanRedlBrack(stringFromStmtref(tmp)), ", ")[[1]]
                if (length(tmp) == 3) out[i,] <- tmp
                else out[i,] <- c(tmp[1:2], paste(tmp[-c(1:2)],collapse=", ")) #assumes 1:2 atomic(URI)
                }
        data.frame(out)
}

stringFromStmtref <- function(stmtref) {
        .Call("string_from_stmtref", stmtref, PACKAGE="Rredland")
}

# ---- graph manipulations

redlMod2graph <- function(x) {
 dd <- lapply( as(x, "data.frame"), as.character )
 no <- unique( c( dd$subject, dd$object ) )
 X <- new("graphNEL", nodes=no, edgemode = "directed" )
 X <- addEdge(dd$subject, dd$object, X )
 edgeDataDefaults(X, "RDFprop") <- ""
 edgeData(X, dd$subject, dd$object, "RDFprop") <- dd[[2]]
 X
}

setAs("redlModel", "graphNEL", function(from, to) redlMod2graph(from) )

# --- node construction

nodeFromURIString = function(string, world=..GredlWorld) {
	new("redlNode", ref=.Call("new_redl_node_from_uri_string",
		ref(world), string), world=world) }
